import json
import cv2
import numpy as np
import logging
from fastapi import FastAPI, WebSocket
from keras.models import load_model
from keras.utils import img_to_array

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize the FastAPI app
app = FastAPI()

# Load the model
emotion_model = load_model('emotion_model.keras')

# Load class indices used during training
with open('class_indices.json', 'r') as f:
    class_indices = json.load(f)

# Convert indices back to labels
emotion_labels = [None] * len(class_indices)
for class_name, index in class_indices.items():
    emotion_labels[index] = class_name

# Load the haarcascade file for face detection
face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_default.xml')

@app.websocket("/ws/emotion")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    try:
        while True:
            # Receive image data from the WebSocket
            data = await websocket.receive_bytes()
            
            # Convert the data to a numpy array
            nparr = np.frombuffer(data, np.uint8)
            frame = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
            
            # Convert color to grayscale
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            # Perform face detection
            faces = face_cascade.detectMultiScale(gray, 1.3, 5)

            emotions = []
            for (x, y, w, h) in faces:
                # Extract grayscale face ROI
                face_gray = gray[y:y + h, x:x + w]

                # Resize the image to 48x48 for model
                face_gray = cv2.resize(face_gray, (48, 48))

                # Convert image to array
                img = img_to_array(face_gray)

                # Reshape image into format for model
                img = np.expand_dims(img, axis=0)

                # Normalize the image
                img /= 255

                # Get the prediction from model
                prediction = emotion_model.predict(img)

                # Get the index of the highest predicted value
                max_index = np.argmax(prediction[0])

                # Get the label corresponding to the prediction
                emotion_prediction = emotion_labels[max_index]

                emotions.append({
                    "emotion": emotion_prediction,
                    "bounding_box": {
                        "x": int(x),
                        "y": int(y),
                        "width": int(w),
                        "height": int(h)
                    }
                })

                logging.info(f"Detected emotions: {emotions}")

            # Send the emotion predictions back to the client
            await websocket.send_text(json.dumps(emotions))

    except Exception as e:
        logging.error(f"An error occurred: {e}")
        await websocket.close()

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)
