import asyncio
import json
import traceback
import cv2
import numpy as np
import logging
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from keras.models import load_model
from keras.utils import img_to_array
import urllib.request
from urllib.error import URLError

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize the FastAPI app
app = FastAPI()

# Load the model
emotion_model = load_model('emotion_model.keras')

# Load class indices used during training
with open('class_indices.json', 'r') as f:
    class_indices = json.load(f)

# Convert indices back to labels
emotion_labels = [None] * len(class_indices)
for class_name, index in class_indices.items():
    emotion_labels[index] = class_name

# Load the haarcascade file for face detection
face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_default.xml')

# Replace with your IP camera stream URL
base_dev_url = 'http://192.168.1.7/'
img_endpoint = '/cam-hi.jpg'

@app.websocket("/ws/emotion/{device_uid}")
async def websocket_endpoint(websocket: WebSocket, device_uid: str):
    await websocket.accept()
    url = base_dev_url+device_uid+img_endpoint
    try:
        while True:
            try:
                # Capture frame from the IP camera stream
                img_resp = urllib.request.urlopen(url)
                imgnp = np.array(bytearray(img_resp.read()), dtype=np.uint8)
                frame = cv2.imdecode(imgnp, cv2.IMREAD_COLOR)

                # Flip the image vertically and horizontally (flipCode=-1)
                flipped_frame = cv2.flip(frame, -1)

                # Convert color to grayscale
                gray = cv2.cvtColor(flipped_frame, cv2.COLOR_BGR2GRAY)

                # Perform face detection
                faces = face_cascade.detectMultiScale(gray, 1.3, 5)

                emotions = []
                for (x, y, w, h) in faces:
                    # Extract grayscale face ROI
                    face_gray = gray[y:y + h, x:x + w]

                    # Resize the image to 48x48 for the model
                    face_gray = cv2.resize(face_gray, (48, 48))

                    # Convert image to array
                    img = img_to_array(face_gray)

                    # Reshape image into format for model
                    img = np.expand_dims(img, axis=0)

                    # Normalize the image
                    img /= 255

                    # Get the prediction from the model
                    prediction = emotion_model.predict(img)

                    # Get the index of the highest predicted value
                    max_index = np.argmax(prediction[0])

                    # Get the label corresponding to the prediction
                    emotion_prediction = emotion_labels[max_index]

                    emotions.append({
                        "emotion": emotion_prediction,
                        "bounding_box": {
                            "x": int(x),
                            "y": int(y),
                            "width": int(w),
                            "height": int(h)
                        },
                        "error": False
                    })

                logging.info(f"Detected emotions: {emotions}")

                # Send the emotion predictions back to the client
                await websocket.send_text(json.dumps(emotions))

            except Exception as e:
                logging.error(f"Error processing frame: {e}")
                traceback.print_exc()
                await websocket.close()

            await asyncio.sleep(1)

    except Exception as e:
        logging.error(f"WebSocket error: {e}")
        traceback.print_exc()
        await websocket.close()
    finally:
        logging.info("Closing WebSocket connection")
        await websocket.close()


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)
